################################################################################
# Created       : Martin Foster
# Modified      : 23-Nov-2003
################################################################################
#
# Database - Modules part of Ethereal Realms, designed to aid database 
#            access, querying and programming  
# Copyright (C) 2000-2003  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 204, 817 5TH ST NE
#                 Calgary, Alberta
#                 T2E 3W9
#
################################################################################

package eRealms::Database;					# Package name

use DBI;							# Database indepedent

################################################################################
# Data Members
################################################################################

################################################################################
# Sub-Routines
################################################################################


################################################################################
# CORE
#
# Core sub-routines are those that are required to be used in order to get basic
# functionality.  Such as connect, the constructor and the destructor.

#####################
# Constructor
#
# Allows to class to be exported and thus spawn objects as a result

sub new
{
        my $class = shift;
        my $self  = { };
        
	# Bless $self into designated class
	bless $self, $class;

	# Return self	
        return $self;
}


#####################
# Database Connect
#
# Sub-routine to allow connection to the database while revealing as little as
# possible to the coder.  Also centralized

sub Connect
{
	#####################
	# Data members

	my $self     = shift;					# Self reference
	my $database;				 		# Database handle

	# User variables
	my $dbname	= 'ethereal';				# Database
	my $host	= 'CHANGE';				# Server to connect
	my $port	= '5432';				# Port to use
	my $user	= 'ethereal';				# Login name
	my $pass	= 'CHANGE';				# Password
	

	#####################
	# Program area

	# Establish connection
	$database = DBI->connect("DBI:Pg:dbname=$dbname;
			host=$host;
			port=$port;
			user=$user;
			password=$pass"
	 ) || die ("Database error \#$DBI::err :  $DBI::errstr");

	# Default behaviour
	$database->{AutoCommit} = 0;
	$database->{RaiseError} = 1;

	# Store handle
	$self->{'HANDLE'} = $database;
}


#####################
# Destructor
#
# Subroutine that will kick in once the object itself has passed away.  This
# will allow to automatically kill the database connection and this free
# us of adding that code

sub DESTROY
{
	#####################
	# Data Members

	my $self     = shift;					# Reference to self
	my $database = $self->{'HANDLE'};			# Database handle

	
	#####################
	# Program area

	# Commit to remove any shared locks
	$database->commit();

	# Connection pooling correction
	$database->{AutoCommit} = 1;

	# Disconnect from database
	$database->disconnect();
}


#####################
# Rollback
#
# A smple wrapper for the DBI commit method.

sub Commit
{
	#####################
	# Data Members

	my $self     = shift;					# Reference to self
	my $data     = $self->{'HANDLE'};			# Database handle

	
	#####################
	# Program area

	# Commit changes
	$data->commit();

}


#####################
# Rollback
#
# A smple wrapper for the DBI rollback method.

sub Rollback
{
	#####################
	# Data Members

	my $self     = shift;					# Reference to self
	my $data     = $self->{'HANDLE'};			# Database handle

	
	#####################
	# Program area

	# Rollback changes
	$data->rollback();
}


################################################################################
# STATIC
#
# Static sub-routines exist to retreive very specific information based on a few
# parameters passed to it.   This essentially means that it does most of the
# work and already has the queries built in.


#####################
# Default Puppet
#
# While this makes use of Data Get, the use of this function is very much
# static.

sub DataGetDefault
{ 
	#####################
	# Data members
	my $self     = shift;					# Reference to self
	my $user     = shift;					# Users login name
	my $realm    = shift;					# Realm name
	my $access   = shift;					# Access control

	my $def_pup;						# Default puppet 
	my $def_realm;						# Default realm


	#####################
	# Program area

	# Puppets
	# Non canon handles
	unless ((defined($access)) && ($access eq 'restricted'))
	{
		# Realm pool
		$realm = 'system';

		# Selected default
		($def_pup) = $self->DataGet("SELECT PuppetName
			FROM Puppet
			WHERE PuppetDefault=?
			  AND PuppeteerLogin=?", 'yes', $user);

		# No default specified
		unless (defined($def_pup))
		{
			# Alphebetical search
			($def_pup) = $self->DataGet("SELECT PuppetName 
				FROM Puppet 
				WHERE PuppetPool=? 
				  AND PuppeteerLogin=? 
				ORDER BY PuppetName LIMIT 1",  $realm, $user); 
		}
	}

	# Canon handles
	else
	{
		# Alphebetical search
		($def_pup) = $self->DataGet("SELECT PuppetName
			FROM Puppet
			WHERE PuppetPool=?
			AND PuppeteerLogin=?
			ORDER BY PuppetName LIMIT 1", $realm, $user);
	}


	# Puppeteer
	# Initial pull
	($def_realm) = $self->DataGet("SELECT PuppeteerRealm
		FROM Puppeteer
		WHERE PuppeteerLogin=?", $user);

	# Does not exist
	unless (defined($def_realm))
	{
		# Auxilary default
		($def_realm) = $self->DataGet("SELECT RealmName
			FROM Realm
			WHERE RealmPublic=?
			ORDER BY RealmName LIMIT 1", 'yes');
	}


	# Returns default handle
	return $def_pup, $def_realm;
}


#####################
# Work Average
#
# This allows the gallery to make proper calculations on the ratings of specific
# works and designs.

sub DataGetWorkAvg
{
	#####################
	# Data members
	my $self     = shift;					# Reference to self
	my $gid      = shift;					# Gallery ID
	my $wid      = shift;					# WorkID

	my $avg;						# Average
	my $vote;						# Votes


	#####################
	# Program area

	# Retrieve
	($avg, $vote) = $self->DataGet("SELECT
		  ROUND(AVG(CommentRating)),
		  COUNT(CommentRating)
		FROM Comment
		WHERE GalleryID=?
		AND WorkID=?
		AND CommentRating <> 0", 

	 $gid,
	 $wid);

	# Ensure assignment
	$avg  = (defined($avg))  ? $avg  : 0;
	$vote = (defined($vote)) ? $vote : 0;

	# Retrun
	return $avg, $vote;
}


#####################
# Group List
#
# Group list exists to generate a list containing the groups and puppets
# for in chat use.

sub GroupList
{
        #####################
        # Data members
	my $self     = shift;					# Self referencing link
	my $list     = shift;					# Reference to array
	my $access   = shift;					# Access method
	my $user     = shift;					# User name
	my $room     = shift;					# Room name

	my $statement;						# Statement handle

	my $current;						# Current group
	my $previous = 'NIL';					# Previous group

	my $plan;						# Plan of execution
	my $puppet;						# Puppet selected


        #####################
	# Program Area

	# What to do
	$plan = ($access ne 'restricted') ? "(PuppetPool='system' OR PuppetPool=?)" : "PuppetPool=?";

	# Get things started
	$self->Pull(\$statement, "SELECT
		  PuppetAssociation,
		  PuppetName
		FROM Puppet
		WHERE PuppeteerLogin=?
		  AND PuppetActive IS NULL
		  AND $plan
		ORDER BY PuppetAssociation, LOWER(PuppetName)",
	 $user,
	 $room
	);

	# Loop and analyse
	# Groups planned
	while (($current, $puppet) = $statement->fetchrow())
	{
		# Checks
		# Leave if not needed
		last unless (defined($previous));
		last unless (defined($current));

		# Handling
		# Titles
		if ($current ne $previous)
		{
			# Add a space
			push(@{$list}, '') if ($previous ne 'NIL');

			# Add title
			push(@{$list}, "-- $current --");

			# Make them match
			$previous = $current;
		}
		
		# General insert
		push(@{$list}, $puppet);
	}

	# Verify if there is still something there
	if (defined($puppet))
	{
		# Additional check
		push(@{$list}, '') if ((defined($previous)) && ($previous ne 'NIL'));	

		# Loop and fetch the rest
		do
		{
			# Just put on the rest of the list
			push(@{$list}, $puppet);
		} while (($current, $puppet) = $statement->fetchrow());
	}

	# Finish query
	$statement->finish();
}


#####################
# Document Get
#
# Data retreival sub-routine for tag templates

sub DocumentGet	
{
	#####################
	# Data members

	my $self     = shift;					# Referene to self
	my $field    = shift;					# Search field

	my $data;						# Data pulled from database
	my $database = $self->{'HANDLE'};			# Database handle
	my $statement;						# Query statement

	my $type     = 'template';				# Type of tag


	#####################
	# Program area

	# Initialize and execute
	$statement = $database->prepare("SELECT TagData 
		FROM  Tag 
		WHERE TagName=?
		  AND TagType=?");
	$statement->execute($field, $type);

	# Retreive and finish
	$data = $statement->fetchrow();
	$statement->finish;

	# Return gathered information
	return $data;
}
 
# Alias for Acceptable Use Policy
sub DocumentGetAUP
{ print DocumentGet(shift, 'WebPolicy'); }

# Banner administration
sub DocumentGetBanner
{ return DocumentGet(shift, 'WebBanner'); }

# Board admin
sub DocumentGetBoardAdmin
{ return DocumentGet(shift, 'BoardAdmin'); }

# Board notice
sub DocumentGetBoardNotice
{ return DocumentGet(shift, 'BoardNotice'); }

# Bio base
sub DocumentGetBioBase
{ return DocumentGet(shift, 'BioBase'); }

# Bio private message
sub DocumentGetBioPrivate
{ return DocumentGet(shift, 'BioPrivate'); }

# Canon handle realm
sub DocumentGetCanon
{ print DocumentGet(shift, 'WebCanon'); }

# Cookie management page
sub DocumentGetCookie
{ return DocumentGet(shift, 'WebCookie'); }

# Donations section
sub DocumentGetDonations
{ return DocumentGet(shift, 'WebDonation'); }

# Donations Mail
sub DocumentGetDonationsMail
{ return DocumentGet(shift, 'MailDonation'); }

# Alias for Gallery Header
sub DocumentGetGalleryHeader
{ print DocumentGet(shift, 'GalleryHeader'); }

# Alias for Gallery Footer
sub DocumentGetGalleryFooter
{ print DocumentGet(shift, 'GalleryFooter'); }

# Alias for Gallery Notice
sub DocumentGetGalleryNotice
{ return DocumentGet(shift, 'GalleryNotice'); }

# Alias for Gallery Search
sub DocumentGetGallerySearch
{ return DocumentGet(shift, 'GallerySearch'); }

# Alias for Gallery Type
sub DocumentGetGalleryType
{ return DocumentGet(shift, 'GalleryType'); }

# Alias for Header
sub DocumentGetHeader
{ print DocumentGet(shift, 'WebHeader'); }

# Alias for Header
sub DocumentGetHeaderPlain
{ print DocumentGet(shift, 'WebHeaderPlain'); }

# Alias for Home Header
sub DocumentGetHomeHeader
{ return DocumentGet(shift, 'HomeHeader'); }

# Alias for Home Readme
sub DocumentGetHomeReadme
{ return DocumentGet(shift, 'HomeReadme'); }

# Alias for Footer
sub DocumentGetFooter
{ print DocumentGet(shift, 'WebFooter'); }

# Alias for Invitation
sub DocumentGetInvitation
{ return DocumentGet(shift, 'MailInvitation'); }

# Mail banned
sub DocumentGetMailBanned
{ return DocumentGet(shift, 'MailBanned'); }

# Mail gallery
sub DocumentGetMailGallery
{ return DocumentGet(shift, 'MailGallery'); }

# Mail notice
sub DocumentGetMailNotice
{ return DocumentGet(shift, 'MailNotice'); }

# Mail posts notificiation and instructions
sub DocumentGetMailPosts
{ return DocumentGet(shift, 'MailPosts'); }

# Mail confirmed
sub DocumentGetConfirmed
{ return DocumentGet(shift, 'MailVerified'); }

# Alias for Confirmation
sub DocumentGetConfirmation
{ print DocumentGet(shift, 'WebConfirmation'); }

# Alias for Description handler
sub DocumentGetDescription
{ return DocumentGet(shift, 'WebDescription'); }

# Alias for Registration
sub DocumentGetRegistration
{ print DocumentGet(shift, 'WebRegistration'); }

# Alias for Accepted registration
sub DocumentGetAccepted
{ print DocumentGet(shift, 'WebAccepted'); }

# Alias for Javascript
sub DocumentGetJavascript
{ print DocumentGet(shift, 'WebJavascript'); }

# Alias for Webring
sub DocumentGetWebring
{ return DocumentGet(shift, 'WebRing'); }

# Alias for Stream Kill
sub DocumentGetStream
{ return DocumentGet(shift, 'WebStream'); }

# Alias for Stream Kill
sub DocumentGetReset
{ return DocumentGet(shift, 'MailReset'); }


#####################
# Fetch Realm
#
# Used to retrieve the information from realms, making first sure that they
# have indeed made use of the functionality.  Otherwise things, will tend
# to falter.

sub FetchRealm
{
	#####################
	# Data members

	my $self     = shift;					# Referene to self
	my $field    = shift;					# Search field
	my $realm    = shift;					# Affected realm

	my $value;						# Value pulled from database


	#####################
	# Program area

	# Initial retrieval
	($value) = $self->DataGet("SELECT DesignContent
		FROM  RealmDesign
		WHERE RealmName=?
		  AND TagName=?", $realm, $field);


	# If no value exists
	unless (defined($value))
	{
		# Retreive defaults
		$value = $self->DocumentGet($field);
	}

	# Return value for use
	return $value;
}

# Alias for dice roller
sub FetchRealmDice
{ return FetchRealm(shift, 'RealmDice', shift); }

# Alias for footer
sub FetchRealmFooter
{ return FetchRealm(shift, 'RealmFooter', shift); }

# Alias for header 
sub FetchRealmHeader
{ return FetchRealm(shift, 'RealmHeader', shift); }

# Alias for information
sub FetchRealmInfo
{ return FetchRealm(shift, 'RealmInfo', shift); }

# Alias for javascript
sub FetchRealmJavascript
{ return FetchRealm(shift, 'RealmJavascript', shift); }

# Alias for notices
sub FetchRealmNotice
{ return FetchRealm(shift, 'RealmNotice', shift); }

# Alias for panel
sub FetchRealmPanel
{ return FetchRealm(shift, 'RealmPanel', shift); }

# Alias for private messages
sub FetchRealmPrivate
{ return FetchRealm(shift, 'RealmPrivate', shift); }

# Alias for Refresh frame
sub FetchRealmRefresh
{ return FetchRealm(shift, 'RealmRefresh', shift); }

# Alias for auto-scrolling remote control
sub FetchRealmRemote
{ return FetchRealm(shift, 'RealmRemote', shift); }

# Alias for seperator bar
sub FetchRealmSeperator
{ return FetchRealm(shift, 'RealmSeperator', shift); }

# Alias for supplementary panel code
sub FetchRealmSupp
{ return FetchRealm(shift, 'RealmSupp', shift); }

# Alias for warnings
sub FetchRealmWarning
{ return FetchRealm(shift, 'RealmWarning', shift); }

# Alias for zoom utility
sub FetchRealmZoom
{ return FetchRealm(shift, 'RealmZoom', shift); }


#####################
# Fetch Banner
#
# This is used to fetch banners for either the system or realm specific.

sub FetchBanner
{ 
	#####################
	# Data members
	my $self     = shift;					# Referene to self
	my $sub      = shift;					# Subtype return

	my $banner;						# Retreived banned
	my $data;						# Fortune data
	my $idnt;						# Fortune type


	#####################
	# Program area
	
	# Fetch banner
	$banner = ($sub != 1) ? $self->FetchRealm('RealmBanner', shift) : $self->DocumentGetBanner();

	
	# Verify for fortune
	if ($banner =~ /FRTN([\w\-]+)/s)
	{
		# Force case
		$idnt = lc($1);

		# Retreive data
		($data) = $self->DataGetFortune($idnt);


		# Replace if defined
		if (defined($data))
		{
			# Use $1 for proper case
			$banner =~ s/FRTN$1/$data/gs;
		}
	}


	# Return banner
	return $banner;
}

# Realm specific banners
sub FetchRealmBanner
{ return FetchBanner(shift, 0, shift); }

# System banners
sub FetchWebBanner
{ return FetchBanner(shift, 1); }


#####################
# Realm formatting
#
# This will return a date/time stamp with the formatting as specified
# by the hosts.

sub GetDateFormat
{
	#####################
	# Data members
	my $self     = shift;					# Reference to self
	my $realm    = shift;					# Realm name query

	my $format;						# Format style
	my $stamp;						# Actual timestamp

	#####################
	# Program area

	# Pull format
	($format) = $self->DataGet("SELECT RealmDateFormat
		FROM Realm
		WHERE RealmName=?", $realm);

	# Fetch formatted date
	($stamp)  = $self->DataGet("SELECT TO_CHAR(NOW(), ?) AS Realm_Date", $format);


	# Return
	return $stamp;
} 


#####################
# Get Hashes
#
# Sub-routine created to take a reference to an array and retrieve a list of
# all records within a specific field of information.

sub GetHash 
{
	#####################
	# Data members
	my $self     = shift;					# Reference to self
	my $hash     = shift;					# Reference of hash
	my $type     = shift;					# Type of values

	my $database = $self->{'HANDLE'};			# Database handle
	my $statement;						# SQL Statement

	my $key;						# Key of hash
	my $value;						# Value for hash


	#####################
	# Program area

	# Prepare and execute
	$statement = $database->prepare("SELECT 
		 TagName, 
		 TagData
		FROM  Tag
		WHERE TagType=?");
	$statement->execute($type);

	# Retreive and use
	while (($key, $value) = $statement->fetchrow())
	{ $hash->{$key} = $value; }

	# Finish statement
	$statement->finish();
}

# Alias for Administrative hash
sub GetHashAdmin
{ GetHash(shift, shift, 'admin'); }

# Alias for Board hash
sub GetHashBoard
{ GetHash(shift, shift, 'board'); }

# Alias for Configuration hash
sub GetHashConfiguration
{ GetHash(shift, shift, 'configuration'); }

# Alias for Error codes
sub GetHashErr
{ GetHash(shift, shift, 'vcodes'); }

# Alias for Gallery Hash
sub GetHashGallery
{ GetHash(shift, shift, 'gallery'); }

# Alias for Chat hash
sub GetHashHome
{ GetHash(shift, shift, 'home'); }

# Alias for Login hash
sub GetHashLogin
{ GetHash(shift, shift, 'login'); }

# Alias for Mail hash
sub GetHashMail
{ GetHash(shift, shift, 'mail'); }

# Alias for Poll hash
sub GetHashPoll
{ GetHash(shift, shift, 'poll'); }

# Alias for Portal hash
sub GetHashPortal
{ GetHash(shift, shift, 'portal'); }

# Alias for Realm hash
sub GetHashRealm
{ GetHash(shift, shift, 'realm'); }

# Alias for System hash
sub GetHashSystem
{ GetHash(shift, shift, 'system'); }


################################################################################
# DYNAMIC
#
# Dynamic sub-routines perform a specific task, such as return a row or a list
# but does not know the formation of the query until it is passed to it.

#####################
# GetList
#
# Sub-routine created to take a reference to an array and retrieve a list of
# all records within a specific field of information.

sub GetList
{
	#####################
	# Data members
	my $self     = shift;					# Reference to self
	my $array    = shift;					# Reference of array
	my $query    = shift;					# SQL query

	my $database = $self->{'HANDLE'};			# Database handle
	my $statement;						# SQL Statement


	#####################
	# Program area

	# Prepare and execute
	$statement = $database->prepare($query);
	$statement->execute(@_);

	# Cleanup of array
	splice(@{ $array }, 0);

	# Retrieve and place in array
	while ($_ = $statement->fetchrow())
	{ push(@{ $array }, $_); }

        # End statement
        $statement->finish;
}

# Alias for designs
sub GetListDesign
{
  GetList(shift, shift,
 	"SELECT TagName
	 FROM Tag
	 WHERE TagName LIKE 'Realm%'
	   AND TagType=?", 'template');
}

# Alias for documents
sub GetListDocuments
{
  GetList(shift, shift, 
	"SELECT TagName
	 FROM Tag 
	 WHERE TagType=? 
	 ORDER BY TagName", 'template');
}

# Alias for estavblished genres
sub GetListEstablished
{
  GetList(shift, shift,
	"SELECT GenreName
	 FROM GalleryGenre
	 WHERE GalleryID=?
	 ORDER BY GenreName", @_);
}

# Alias for genres
sub GetListGenre
{ 
  GetList(shift, shift, 
	"SELECT GenreName
	 FROM Genre
	 WHERE GenreOrder=?
	 ORDER BY GenreName", @_); 
}

# Alias for genres
sub GetListGroup
{ 
  GetList(shift, shift, 
	"SELECT AssociationName
	 FROM Association
	 WHERE PuppeteerLogin=?
	 ORDER BY AssociationName", @_); 
}

# Alias for public realms
sub GetListPublicRealms
{
  GetList(shift, shift, 
	"SELECT RealmName
	 FROM Realm
	 WHERE RealmPublic=?
	 ORDER BY RealmName", 'yes'); 
}

# Alias for work sort worder
sub GetListWorkSort
{
  GetList(shift, shift,
	"SELECT WorkID
	 FROM Work
	 WHERE GalleryID=?
	 ORDER BY WorkSort", @_); 
}

# Alias for private realms
sub GetListSupervisorRealms
{
  GetList(shift, shift,
	"SELECT RealmName
	 FROM RealmSupervisor
	 WHERE PuppeteerLogin=?
	 ORDER BY RealmName", @_); 
}

# Alias for non-private realms
sub GetListNonPrivate
{
  GetList(shift, shift,
	"SELECT RealmName
	 FROM Realm
	 WHERE (RealmPublic=?
	    OR RealmPublic=?)
	   AND RealmName NOT IN ('system', 'banned')
	 ORDER BY RealmName", 'yes', 'hybrid'); }

# Alias for non-private realms
sub GetListVerifiedRealms
{
  GetList(shift, shift,
	"SELECT RealmName
	 FROM Realm
	 WHERE RealmPublic <> ?
	 AND RealmVerified=?
	 ORDER BY RealmName", 'yes', 'yes'); 
}

# System and Realm handles
sub GetListPuppetRealm
{
  GetList(shift, shift, 
	"SELECT PuppetName
	 FROM Puppet
	 WHERE PuppeteerLogin=?
	   AND PuppetActive IS NULL
	   AND (PuppetPool=?
	    OR PuppetPool=?)
	 ORDER BY PuppetName", shift, 'system', @_);
}

# Alias for complete puppet listing
sub GetListPuppetComplete
{ 
  GetList(shift, shift, 
	"SELECT PuppetName
	 FROM Puppet
	 WHERE PuppeteerLogin=?
	 ORDER BY PuppetName", @_); 
}

# Alias for active puppet listing that are yours
sub GetListPuppetYours
{
  GetList(shift, shift, 
	"SELECT PuppetName
	 FROM Puppet
	 WHERE PuppetPool=?
	   AND PuppeteerLogin=?
	   AND PuppetActive IS NULL
	 ORDER BY PuppetName", 'system', @_); 
}

# Alias for puppet listing that are yours
sub GetListPuppetSuspended
{
  GetList(shift, shift, 
	"SELECT PuppetName
	 FROM Puppet
	 WHERE PuppeteerLogin=?
	 AND PuppetActive IS NOT NULL
	 ORDER BY PuppetName", @_); 
}

# Alias for EMail retreival
sub GetListEMails
{
  GetList(shift, shift, 
	"SELECT p.PuppeteerEMail
	 FROM Puppeteer p,
	      PuppeteerRegistration r
	 WHERE p.PuppeteerLogin <> r.PuppeteerLogin
	 GROUP BY PuppeteerEMail
	 ORDER BY PuppeteerEMail"); 
}


# Canon realm exceptions
# Alias for Canon handles
sub GetListCanonPuppets
{
  GetList(shift, shift, 
	"SELECT PuppetName
	 FROM Puppet
	 WHERE PuppeteerLogin=?
	   AND PuppetActive IS NULL
	   AND PuppetPool=?
	 ORDER BY PuppetName", @_); 
}

# Alias for locale
sub GetListRealmLocale
{
  GetList(shift, shift, 
	"SELECT RealmLocale
	 FROM RealmLocale
	 WHERE RealmName=?
	 ORDER BY RealmLocale", @_); 
}

# Restrictions
# Alias for restricted words
sub GetListWords
{  
  GetList(shift, shift, 
	"SELECT RestrictionValue
	 FROM Restriction
	 WHERE RealmName=?
	   AND RestrictionType=?
	 ORDER BY RestrictionValue", shift, 'word', @_); 
}

# Banned images
sub GetListImageBan
{
  GetList(shift, shift, 
	"SELECT RestrictionValue
	 FROM Restriction
	 WHERE RealmName=?
	   AND RestrictionType=?", 'system', 'picture', @_); 
}

# Board Topics
sub GetListTopics
{
  GetList(shift, shift, 
	"SELECT BoardTopic
	 FROM BoardTopic
	 WHERE RealmName=?
	 ORDER BY BoardTopic", @_); 
}


#####################
# GetData
#
# Generic sub-routine that will return a single row of data be it in multiple
# amounts or just one field within the row there is a special section for
# font handling if choice is at all defined.

sub DataGet
{
	#####################
	# Data members
	my $self     = shift;					# Reference to self
	my $query    = shift;					# Database query

	my $database = $self->{'HANDLE'};			# Database get
	my $statement;						# SQL Statement

	my @values;						# Values returned from table


	#####################
	# Program area

	# Prepare and execute
	$statement = $database->prepare($query);
	$statement->execute(@_);

	# Retreive and finish
	@values = $statement->fetchrow();
	$statement->finish();

	# Return retreived information
	return @values;
}


# Boards
# Alias for administrative Email
sub DataGetRealmBoard
{
  return DataGet(shift, 
	"SELECT RealmBoard
	 FROM Realm
	 WHERE RealmName=?",  @_); 
}


# Dates
# Alias for short date
sub DataGetDate
{
  return DataGet(shift, "SELECT getDate(NOW()) AS Short_Date", @_);
}

# Alias for full timestamp
sub DataGetDateTime
{
  return DataGet(shift, "SELECT getDateTime(NOW()) AS Long_Date", @_);
}

# Frequency
sub DataGetFrenquency
{
  return DataGet(shift,
	"SELECT COUNT(*)
	 FROM Puppeteer
	 WHERE PuppeteerTimestamp > getInterval('" . shift(@_) . " day')", @_);
}


# Gummies
# Alias for class lookup
sub DataGetGummyClass
{
  return DataGet(shift, 
	"SELECT GummyClass
	 FROM Gummy
	 WHERE GummyPath=?",  @_); 
}

# Alias for name lookup
sub DataGetGummyName
{
  return DataGet(shift, 
	"SELECT GummyName
	 FROM Gummy
 	 WHERE GummyPath=?",  @_); 
}

# Alias for pathname lookup
sub DataGetGummyPath
{
  return DataGet(shift, 
	"SELECT GummyPath
	 FROM Gummy
	 WHERE GummyName=?",  @_); 
}


# Parameter list
# Parameter line
sub DataGetParam
{
  return DataGet(shift, 
	"SELECT ParamValue
	 FROM Param
	 WHERE PuppeteerLogin=?
	   AND ParamNumber=?::INT", @_); 
}

# Sequence number
sub DataGetSequence
{
  return DataGet(shift, 
	"SELECT ParamNumber
	 FROM Param
	 WHERE PuppeteerLogin=?
	 AND ParamNumber=?::INT", @_); 
}


# Privileges
# Alias for Administrative  option
sub DataGetAdmin
{
  return DataGet(shift, 
	"SELECT PuppeteerAdminPrivs
	 FROM Puppeteer
	 WHERE PuppeteerLogin=?", @_); 
}

# Alias for Supervisory Option
sub DataGetSuper
{ 
  return DataGet(shift, 
	"SELECT PuppeteerSuperPrivs
	 FROM Puppeteer
	 WHERE PuppeteerLogin=?", @_); 
}

# Alias for Privileges
sub DataGetPrivs
{
  return DataGet(shift, 
	"SELECT PuppeteerSuperPrivs,
	        PuppeteerAdminPrivs,
	        PuppeteerModPrivs, 
	        PuppeteerVerified
	 FROM Puppeteer
	 WHERE PuppeteerLogin=?", @_);
}

# Realm privs
sub DataGetRoomPrivs
{
  return DataGet(shift, 
	"SELECT p.PuppeteerLogin
	 FROM Puppeteer p,
	      RealmSupervisor r
	 WHERE p.PuppeteerLogin=r.PuppeteerLogin
	   AND r.RealmName=?
	   AND p.PuppeteerLogin=?
	   AND p.PuppeteerSuperPrivs=?", shift, shift, 'yes', @_); }

# Moderator counts
sub DataGetModCount
{
  return DataGet(shift, 
	"SELECT COUNT(*)
	 FROM RealmSupervisor
	 WHERE PuppeteerLogin=?", @_);
}


# Posts
# Post ID Number
sub DataGetPostID
{
  return DataGet(shift, 
	"SELECT PostIDNumber
	 FROM Post
	 WHERE PostTimeStamp > getInterval('10 Minutes') LIMIT 1", @_);
}

# Puppeteer
# Alias for Puppeteer Info
sub DataGetPuppeteerLogin
{
  return DataGet(shift,
	"SELECT PuppeteerLogin
	 FROM Puppet
	 WHERE PuppetPool=?
	   AND LOWER(PuppetName)=LOWER(?)", 'system', @_); 
}

# Alias for canon handles
sub DataGetPuppeteerCanon
{ 
  return DataGet(shift, 
	"SELECT PuppeteerLogin
	 FROM Puppet
	 WHERE LOWER(PuppetPool)=LOWER(?)
	 AND LOWER(PuppetName)=LOWER(?)", @_); 
}


# Alias for font settings
sub DataGetFont
{
	# Pull values
	my ($text, $color, $face) = DataGet(shift, 
		"SELECT PuppetTextSize,
		        PuppetTextColour,
		        PuppetTextFace
		 FROM Puppet
		 WHERE PuppetName=?
		 AND PuppeteerLogin=?", shift, shift); 

	# Verify and define
	$text  = (defined($text))  ? $text  : 3;
	$color = (defined($color)) ? $color : shift;
	$face  = (defined($face))  ? $face  : shift;

	return ($text, $color, $face);
}

# Alias for Javascript option
sub DataGetJavascript
{
  return DataGet(shift,
	"SELECT PuppeteerJavascript
	 FROM Puppeteer
	 WHERE PuppeteerLogin=?", @_); 
}

# Alias for chat options
sub DataGetChatOptions
{
  return DataGet(shift, 
	"SELECT PuppeteerJavascript,
	        PuppeteerChatPref
	 FROM Puppeteer
	 WHERE PuppeteerLogin=?", @_); 
}

# Alias for Formatting option
sub DataGetFormatting
{
  return DataGet(shift, 
	"SELECT PuppeteerFormatting,
                PuppeteerChatPref,
                PuppeteerPrivacy 
	 FROM Puppeteer
	 WHERE PuppeteerLogin=?", @_); 
}

# Alias for Ignore Scope
sub DataGetScope
{
  return DataGet(shift,
	"SELECT PuppeteerIgnore
	 FROM Puppeteer
	 WHERE PuppeteerLogin=?",  @_);
}

# Alias for Supervisory option
sub DataGetBannedStatus
{
  return DataGet(shift, 
	"SELECT PuppeteerLogin
	 FROM PuppeteerRestriction
	 WHERE PuppeteerLogin=?
	   AND RealmName=?", @_); 
}


# Alias for Puppeteers Control over a realm
sub DataGetPuppeteerControl
{
  return DataGet(shift,
	"SELECT PuppeteerLogin
	 FROM  RealmSupervisor
	 WHERE RealmName=?
	  AND PuppeteerLogin=?", @_);
}

# Alias for PuppeteerEmail retreival
sub DataGetPuppeteerEmail
{
  return DataGet(shift, 
	"SELECT PuppeteerEmail
	 FROM Puppeteer
	 WHERE PuppeteerLogin=?", @_); 
}

# Alias for Puppeteers Email
sub DataGetPuppeteerExistence
{
  return DataGet(shift, 
	"SELECT PuppeteerLogin, PuppeteerPassword
	 FROM Puppeteer
	 WHERE LOWER(PuppeteerLogin)=LOWER(?)", @_); 
}

# Alias for Puppeteers Homepage Expire Time
sub DataGetPuppeteerExpire
{
  return DataGet(shift, 
	"SELECT getExpire(?)", @_);
}

# Alias for Puppeteers Homepage
sub DataGetPuppeteerHome
{
  return DataGet(shift,
	"SELECT PuppeteerHome
	 FROM Puppeteer
	 WHERE PuppeteerLogin=?", @_); 
}

# Alias for Mail retreival
sub DataGetPuppeteerMail
{
  return DataGet(shift, 
	"SELECT PuppeteerLogin,
	        MailFrom,
                MailRcpt,
                MailSubject,
                getDateTime(MailTimestamp),
                MailContents
	 FROM Mail
	 WHERE MailIDNumber=?", @_); 
}

# Alias for Mail retreival
sub DataGetPuppeteerMailCount
{
  return DataGet(shift, 
	"SELECT COUNT(*)
	 FROM Mail
	 WHERE PuppeteerLogin=?", @_); 
}

# Alias for Name retreival
sub DataGetPuppeteerName
{
  return DataGet(shift, 
	"SELECT PuppeteerName
	 FROM Puppeteer
	 WHERE PuppeteerLogin=?", @_); 
}

# Alias for registration
sub DataGetPuppeteerReg
{ 
  return DataGet(shift, 
	"SELECT PuppeteerLogin
	 FROM PuppeteerRegistration
	 WHERE PuppeteerLogin=?", @_);
}

# Alias for verified
sub DataGetVerified
{
  return DataGet(shift, 
	"SELECT PuppeteerVerified
	 FROM Puppeteer
	 WHERE PuppeteerLogin=?", @_); 
}

# Puppets
# Active
sub DataGetPuppetActive
{
  return DataGet(shift, 
	"SELECT PuppetActive
	 FROM Puppet
	 WHERE PuppetName=?
	   AND PuppeteerLogin=?", @_); 
}

# Existence
sub DataGetPuppetExistence
{
  return DataGet(shift, 
	"SELECT PuppetName
	 FROM Puppet
	 WHERE LOWER(PuppetName)=LOWER(?)
	 AND PuppeteerLogin=?", @_); 
}

# Alias for puppets locale settings
sub DataGetPuppetLocale
{
  return DataGet(shift, 
	"SELECT PuppetLocale
	 FROM Puppet
	 WHERE PuppetName=?
	   AND PuppeteerLogin=?", @_); 
}

# Alias for title and tagline
sub DataGetPuppetFormat
{
  return DataGet(shift,
	"SELECT PuppetTitle,
	        PuppetTag
	 FROM Puppet
	 WHERE PuppetName=? 
	   AND PuppeteerLogin=?", @_); 
}


# Realm
# Alias for realm access
sub DataGetRealmAccess
{
  return DataGet(shift, 
	"SELECT RealmAccess,
	        RealmVerified
	 FROM Realm
	 WHERE RealmName=?", @_); 
} 

# Alias for realm access
sub DataGetRealmACL
{
  return DataGet(shift, 
	"SELECT RealmACL
	 FROM RealmSupervisor
	 WHERE Puppeteerlogin=?
	   AND RealmName=?", @_); 
}

# Alias for alternate message boards
sub DataGetRealmAltBoard
{ 
  return DataGet(shift, 
	"SELECT RealmAltBoard
	 FROM Realm
	 WHERE RealmName=?", @_); 
}

# Alias for realm archive address
sub DataGetRealmArchive
{
  return DataGet(shift,
	"SELECT RealmArchive
	 FROM Realm
	 WHERE RealmName=?", @_); 
}

# Alias for realm description
sub DataGetRealmDescription
{
  return DataGet(shift, 
	"SELECT DesignContent
	 FROM RealmDesign
	 WHERE TagName=?
	 AND RealmName=?", 'RealmInfo', @_); 
}

# Alias for Realm existence
sub DataGetRealmExistence
{
  return DataGet(shift,
	"SELECT RealmName,
	        RealmPublic,
	        RealmAlias
	 FROM Realm
	 WHERE LOWER(RealmName)=LOWER(?)", @_); 
}

# Alias for realm fonts
sub DataGetRealmFont
{
  return DataGet(shift, 
	"SELECT RealmFontColor,
	        RealmFontFace,
	        RealmFontSize
	 FROM Realm
	 WHERE RealmName=?", @_); }

# Alias for Frameset info
sub DataGetRealmFrameset
{
  return DataGet(shift, 
	"SELECT RealmFramesetCols,
	        RealmFramesetRows
	 FROM Realm
	 WHERE RealmName=?", @_); 
}

# Alias for Realm Image Info
sub DataGetRealmImageInfo
{
  return DataGet(shift, 
	"SELECT RealmImageHeight,
	        RealmImageWidth,
	        RealmImageSize
	 FROM Realm
	 WHERE RealmName=?", @_); 
}

# Alias for locale settings
sub DataGetRealmLocale
{
  return DataGet(shift, 
	"SELECT RealmLocale
	 FROM Realm
	 WHERE RealmName=?", @_); 
}


# Alias for Realm password
sub DataGetRealmPassword
{
  return DataGet(shift, 
	"SELECT RealmPassword
	 FROM Realm
	 WHERE RealmName=?", @_); 
}

# Alias for Realm Public status
sub DataGetRealmPublic
{
  return DataGet(shift,
	 "SELECT RealmPublic,
	         RealmVerified
	 FROM Realm
	 WHERE RealmName=?", @_); 
}

# Alias for Realm Public status
sub DataGetRealmText
{
  return DataGet(shift, 
	"SELECT RealmTextColour
	 FROM Realm
	 WHERE RealmName=?", @_);
}


# Misc
# Insert ID
sub DataGetInsert
{
  return DataGet(shift,
	"SELECT NEXTVAL(?)", @_);
}


# Weather control
sub DataGetWeather
{
  return DataGet(shift, 
	"SELECT RealmMETAR
	 FROM Realm
	 WHERE RealmName=?", @_); 
}

# Fortunes
sub DataGetFortune
{
  return DataGet(shift, 
	"SELECT FortuneMsg
	 FROM Fortune
	 WHERE FortuneType=?
	 ORDER BY RANDOM() LIMIT 1", @_);
}

# Alias for font code
sub DataGetFontCode
{
  return DataGet(shift, 
	"SELECT FontCode
	 FROM Font
	 WHERE FontName=?
	 AND FontType=?", @_);
}

# Weblog articles
sub DataGetWeblogEntries
{
  return DataGet(shift,
	 "SELECT COUNT(*)
	 FROM Weblog", @_);
}


# Gallery
# Alternate gallery basics
sub DataGetGalleryBasics
{
  return DataGet(shift, 
	"SELECT GalleryName, GalleryPenName
	 FROM Gallery
	 WHERE GalleryID=?", @_); 
}

# Gallery existence
sub DataGetGalleryID
{
  return DataGet(shift, 
	"SELECT GalleryID
	 FROM Gallery
	 WHERE LOWER(GalleryName)=LOWER(?)", @_); 
}

# Alternate Gallery existence
sub DataGetGalleryName
{
  return DataGet(shift, 
	"SELECT GalleryName
	 FROM Gallery
	 WHERE GalleryID=?
	   AND PuppeteerLogin=?", @_); 
}

# Alternate gallery owner
sub DataGetGalleryOwner
{
  return DataGet(shift, 
	"SELECT PuppeteerLogin
	 FROM Gallery
	 WHERE GalleryID=?", @_); 
}

# Gallery Type
sub DataGetGalleryType
{
  return DataGet(shift, 
	"SELECT GalleryType, GallerySort
	 FROM Gallery
	 WHERE GalleryID=?
	 AND PuppeteerLogin=?", @_); 
}

# Gallery Watch
sub DataGetGalleryWatch
{
  return DataGet(shift, 
	"SELECT PuppeteerLogin
	 FROM GalleryNotify
	 WHERE GalleryID=?
	 AND PuppeteerLogin=?", @_); 
}

# Gallery Comments
sub DataGetGalleryComment
{
  return DataGet(shift, 
	"SELECT GalleryComments
	 FROM Gallery
	 WHERE GalleryID=?", @_); 
}

# Work ID
sub DataGetWorkID
{
  return DataGet(shift, 
	"SELECT WorkID
	 FROM Work
	 WHERE LOWER(WorkName)=LOWER(?)
	   AND GalleryID=?", @_); 
}

# Work Name
sub DataGetWorkName
{
  return DataGet(shift, 
	"SELECT WorkName
	 FROM Work
	 WHERE WorkID=?
	   AND GalleryID=?", @_); 
}

# Work Sort
sub DataGetWorkSort
{
  return DataGet(shift, 
	"SELECT WorkSort
	 FROM Work
	 WHERE WorkID=?
	   AND GalleryID=?", @_); 
}


#####################
# Pull
#
# Pull is a wrapper for the complexity behind preparing and retreiving
# information using the prepare and execute handlers.

sub Pull
{
	#####################
	# Data members
	my $self     = shift;					# Reference to self
	my $sth      = shift;					# Query handle
	my $query    = shift;					# Database query

	my $database = $self->{'HANDLE'};			# Database handle


	#####################
	# Program area

	# Prepare query
	${$sth} = $database->prepare($query);

	# Execute
	return ${$sth}->execute(@_);
}


####################
# Query
#
# Query builder, will determin the action to take depending on
# what was recieved in the parameter list.  This will handle
# boolean and simple straight values.

sub Query
{
	####################
	# Data members
	my $self   = shift;				# Reference to self
	my $string = shift;				# String to figure out
	my $field  = shift;				# Field to append to

	my $data   = $self->{'HANDLE'};			# Database handle

	my $choice;					# Various choices
	my $count;					# Counter
	my $logic;					# Logic stream
	my $neg;					# Negative term
	my $size;					# Size of array
	my $word;					# Current word

	my @words;					# Array of dismembered words


	####################
	# Program area


	# Determine behaviour
	# Boolean search
	if ($string =~ /(\(|\)|\sAND\s|\sNOT\s|\sOR\s)/si)
	{
		# Segregate
		@words  = split(/\s/, $string);

		# Reset
		$string = '';

		# Rebuild string
		# Loop
		while ($word = shift(@words))
		{
			# Automatic addition
			if ($word =~ /^(AND|OR)/)
			{
				# Take at face value
				$string = "$string $word";
			}

			# Quotes
			elsif ($word =~ /^('|")(\w+)/)
			{
				# Exact string match
				$string = "$string $field=$word";

				# Find closing
				while (($word = shift(@words))
				 && ($word !~ /(\w+)(\'|\")$/))
				{
					# Add add add
					$string = "$string $word";
				}

				# Append
				$string = "$string $word" if (defined($word));
			}

			else
			{
				# Something at the front
				if ($word =~ /^([\(|\'|\"]+)(\w+)/)
				{ 
					# Quote properly
					$word   = $2;
					$word   = $data->quote("%$word%");
					$string = "$string $1$field $neg ILIKE $word"; 
				}

				# Something in the rear
				elsif ($word =~ /(\w+)([\)|\'|\"]+)$/)
				{
					# Quote properly
					$word   = $1;
					$word   = $data->quote("%$word%");
					$string = "$string $field $neg ILIKE word$2";
				}

				# Standard words
				else
				{
					# Ignore ig a negation
					unless ($word =~ /NOT/i)
					{
						# Quote properly
						$word   = $data->quote("%$word%");
						$string = "$string $field $neg ILIKE $word";
					}
				}


				# Neg negative
				$neg = ($word =~ /NOT/i) ? 'NOT ' : '';
			}
		}
	}	

	# Regular keyword search
	else
	{
		# Segregate
		@words = split(/\s/, $string);

		# Retreive size
		$size  = @words;


		# Reset string
		$string = '';

		# Loop and rebuild
		for ($count=0; $count < $size; $count++)
		{
			# Quotes
			if ($words[$count] =~ /^('|"|\+'|\+"|\-'|\-")(\w+)/)
			{
				# Remove quote
				$words[$count] = "$1$3" if ($words[$count] =~ /^(\+|\-)('|")(\w+)/);
				$words[$count] = $2 if ($words[$count] =~ /^('|")(\w+)/);

				# Find closing
				for (my $icount=$count+1; $icount < $size; $icount++)
				{
					# Default
					unless ($words[$icount] =~ /(\w+)('|")$/)
					{
						# Append work
						$words[$count] = "$words[$count] $words[$icount]";

						# Skip rest 
						next;
					}


					# Found
					# Assign for proper op
					$words[$icount] = "$words[$count] $1";

					# Set position 
					$count = $icount;

					# Last operation
					last;
				}
			}

			# Main handling
			# Exclusion
			$logic  = ($words[$count] =~ /\-\w+/) ? ' NOT ILIKE ' : ' ILIKE ';	

			# Importance
			if ($words[$count] =~ /(\+|\-)(\w)/)
			{
				# Re-assign
				$words[$count] = substr($words[$count], 1);

				# Set logic sequence
				$choice = 'AND';
			}
			else
			{ $choice = 'OR'; }


			# Quote string
			$word = $data->quote("%$words[$count]%");

			# Append to string
			if ($count == 0)
			{ $string = "$field$logic$word"; }
			else
			{ $string = "$string $choice $field$logic$word"; }
		}
	}

	# Return
	return $string;
}

####################
# Quick Write
#
# As the name implies, this sub-routine will write and commit information to
# the database.   This can take place in the form of an update or insert
# statement.  This specific implementation is simple, and is not intended
# for massive transactions.

sub Quick
{
	#####################
	# Data members
	my $self     = shift;					# Reference to self
	my $query    = shift;					# Database query

	my $data     = $self->{'HANDLE'};			# Database handle 


	#####################
	# Program area

	# Evaluation block
	eval
	{
		# Run query
		$data->do($query, {}, @_);
      
		# Commit changes
		$data->commit;
  	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Transaction aborted: $@");

		# Undo incomplete changes
		$data->rollback;
	}
}


####################
# Write
#
# As the name implies, this sub-routine will write and commit information to
# the database.   This can take place in the form of an update or insert
# statement.  This specific requires an external commit statement.

sub Write
{
	#####################
	# Data members
	my $self     = shift;					# Reference to self
	my $query    = shift;					# Database query

	my $data     = $self->{'HANDLE'};			# Database handle 


	#####################
	# Program area

	# Run command
	return $data->do($query, {}, @_);
}

1;								# Return true
